from random import randint
from typing import List
from uuid import uuid4

from orchestrator.db.models import ProductTable, SubscriptionTable
from orchestrator.forms import FormPage
from orchestrator.forms.validators import Choice, choice_list
from orchestrator.targets import Target
from orchestrator.types \
    import FormGenerator, State, SubscriptionLifecycle, UUIDstr
from orchestrator.workflow import done, init, step, workflow
from orchestrator.workflows.steps \
    import resync, set_status, store_process_subscription
from orchestrator.workflows.utils import wrap_create_initial_input_form

from geant_service_orchestrator.products.product_types.trunk_config_common \
    import TrunkConfigCommonInactive, TrunkConfigCommonProvisioning
from geant_service_orchestrator.products.product_types.trunk_config \
    import TrunkConfig


def trunk_config_selector() -> list:
    trunk_config_subscriptions = {}
    for trunk_config_id, trunk_config_description in (
        SubscriptionTable.query.join(ProductTable)
        .filter(
            ProductTable.product_type == "TrunkConfig",
            SubscriptionTable.status == "active",
        )
        .with_entities(
            SubscriptionTable.subscription_id,
            SubscriptionTable.description)
        .all()
    ):
        trunk_config_subscriptions[
            str(trunk_config_id)] = trunk_config_description

    return choice_list(
        Choice("TrunkConfigEnum", zip(
            trunk_config_subscriptions.keys(),
            trunk_config_subscriptions.items())),  # type:ignore
        min_items=1,
        max_items=1,
    )


def initial_input_form_generator(product_name: str) -> FormGenerator:
    class CreateTrunkConfigCommonForm(FormPage):
        class Config:
            title = product_name
        speed: str
        is_leased_line: bool
        isis_metric: int
        minimum_links: int
        trunk_config_ids: trunk_config_selector()  # type:ignore

    user_input = yield CreateTrunkConfigCommonForm
    return user_input.dict()


def _provision_in_gap() -> int:

    return randint(0, 2147483648)


@step("Create subscription")
def create_subscription(product: UUIDstr) -> State:
    subscription \
        = TrunkConfigCommonInactive.from_product_id(product, uuid4())

    return {
        "subscription": subscription,
        "subscription_id": subscription.subscription_id
    }


@step("Initialize subscription")
def initialize_subscription(
        subscription: TrunkConfigCommonInactive,
        speed: str,
        is_leased_line: bool,
        isis_metric: int,
        minimum_links: int,
        trunk_config_ids: List[str]) -> State:
    subscription.trunk_config_common.speed = speed
    subscription.trunk_config_common.is_leased_line = is_leased_line
    subscription.trunk_config_common.isis_metric = isis_metric
    subscription.trunk_config_common.minimum_links = minimum_links
    subscription.trunk_config_common.trunk_config \
        = TrunkConfig.from_subscription(
            trunk_config_ids[0]).trunk_config
    subscription.description = \
        "TrunkConfigCommon from TrunkConfig " \
        f"{subscription.trunk_config_common.trunk_config.trunk_config_id}"
    subscription = TrunkConfigCommonProvisioning.from_other_lifecycle(
        subscription, SubscriptionLifecycle.PROVISIONING)

    return {"subscription": subscription}


@step("Provision trunk config common")
def provision_trunk_config_common(
        subscription: TrunkConfigCommonProvisioning) -> State:
    trunk_config_common_id = _provision_in_gap()
    subscription.trunk_config_common.trunk_config_common_id \
        = trunk_config_common_id

    return {"subscription": subscription}


@workflow(
    "Create trunk config common",
    initial_input_form=wrap_create_initial_input_form(initial_input_form_generator),  # noqa: E501
    target=Target.CREATE,
)
def create_trunk_config_common():

    return (
        init
        >> create_subscription
        >> store_process_subscription(Target.CREATE)
        >> initialize_subscription
        >> provision_trunk_config_common
        >> set_status(SubscriptionLifecycle.ACTIVE)
        >> resync
        >> done
    )
