from typing import List, Optional
from uuid import uuid4

from orchestrator.db.models import ProductTable, SubscriptionTable
from orchestrator.forms import FormPage
from orchestrator.forms.validators import Choice, choice_list
from orchestrator.targets import Target
from orchestrator.types \
    import FormGenerator, State, SubscriptionLifecycle, UUIDstr
from orchestrator.workflow import done, init, step, workflow
from orchestrator.workflows.steps \
    import resync, set_status, store_process_subscription
from orchestrator.workflows.utils import wrap_create_initial_input_form

from geant_service_orchestrator.products.product_types.trunk_config_side \
    import TrunkConfigSideInactive, TrunkConfigSideProvisioning
from geant_service_orchestrator.products.product_types.trunk_config \
    import TrunkConfig


def trunk_config_selector() -> list:
    trunk_config_subscriptions = {}
    for trunk_config_id, trunk_config_description in (
        SubscriptionTable.query.join(ProductTable)
        .filter(
            ProductTable.product_type == "TrunkConfig",
            SubscriptionTable.status == "active",
        )
        .with_entities(
            SubscriptionTable.subscription_id,
            SubscriptionTable.description)
        .all()
    ):
        trunk_config_subscriptions[
            str(trunk_config_id)] = trunk_config_description

    return choice_list(
        Choice("TrunkConfigEnum", zip(
            trunk_config_subscriptions.keys(),
            trunk_config_subscriptions.items())),  # type:ignore
        min_items=1,
        max_items=1,
    )


def member_selector() -> list:
    interfaces_dict = {
        "ge-0/0/1": "ge-0/0/1",
        "ge-0/0/2": "ge-0/0/2",
        "ge-0/0/3": "ge-0/0/3",
        "ge-0/0/4": "ge-0/0/4",
        "ge-0/0/5": "ge-0/0/5"
    }

    return choice_list(
        Choice("Member Interfaces", zip(
            interfaces_dict.keys(),
            interfaces_dict.items())),  # type:ignore
        min_items=1,
        max_items=4,
        unique_items=True,
    )


def initial_input_form_generator(product_name: str) -> FormGenerator:
    class CreateTrunkConfigSideForm(FormPage):
        class Config:
            title = product_name

        fqdn: str
        ae_name: str
        geant_a_sid: str
        ipv4_address: Optional[str]
        ipv6_address: Optional[str]
        members: member_selector()  # type:ignore
        trunk_config_ids: trunk_config_selector()  # type:ignore

    user_input = yield CreateTrunkConfigSideForm

    return user_input.dict()


def _provision_in_gap(fqdn: str) -> int:
    return abs(hash(fqdn))


@step("Create subscription")
def create_subscription(product: UUIDstr) -> State:
    subscription \
        = TrunkConfigSideInactive.from_product_id(product, uuid4())

    return {
        "subscription": subscription,
        "subscription_id": subscription.subscription_id
    }


@step("Initialize subscription")
def initialize_subscription(
        subscription: TrunkConfigSideInactive,
        fqdn: str,
        ae_name: str,
        geant_a_sid: str,
        ipv4_address: str,
        ipv6_address: str,
        members: list,
        trunk_config_ids: List[str]) -> State:
    subscription.trunk_config_side.fqdn = fqdn
    subscription.trunk_config_side.ae_name = ae_name
    subscription.trunk_config_side.geant_a_sid = geant_a_sid
    subscription.trunk_config_side.ipv4_address = ipv4_address
    subscription.trunk_config_side.ipv6_address = ipv6_address
    subscription.trunk_config_side.members = members
    subscription.trunk_config_side.trunk_config \
        = TrunkConfig.from_subscription(trunk_config_ids[0]).trunk_config
    subscription.description = \
        "TrunkConfigSide from TrunkConfig " \
        f"{subscription.trunk_config_side.trunk_config.trunk_config_id}"
    subscription = TrunkConfigSideProvisioning.from_other_lifecycle(
        subscription, SubscriptionLifecycle.PROVISIONING)

    return {"subscription": subscription}


@step("Provision trunk config side")
def provision_trunk_config_side(
        subscription: TrunkConfigSideProvisioning, fqdn: str) -> State:
    trunk_config_side_id = _provision_in_gap(fqdn)
    subscription.trunk_config_side.trunk_config_side_id = trunk_config_side_id

    return {"subscription": subscription}


@workflow(
    "Create trunk config side",
    initial_input_form=wrap_create_initial_input_form(initial_input_form_generator),  # noqa: E501
    target=Target.CREATE,
)
def create_trunk_config_side():

    return (
        init
        >> create_subscription
        >> store_process_subscription(Target.CREATE)
        >> initialize_subscription
        >> provision_trunk_config_side
        >> set_status(SubscriptionLifecycle.ACTIVE)
        >> resync
        >> done
    )
